/***************************************************************************
                          exemples.h  -  description
                             -------------------

    functionality        : for defining and handling (sampling) datasets
    relevant classes     : example, dataset
    relevant functions   : randomGenerator

    begin                : 9 Oct 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef EXEMPLES_H
#define EXEMPLES_H

#include "parser.h"
#include <vector>
#include <string>
#include <set>
#include "rules.h"
#include "estructures.h"
#include <stdio.h>
#include <algorithm>
#include <time.h>

class options;

void ompli_memoria(const string & fitxer,dataset  & dset, const options &Options);
void ompli_memoria(const string & fitxer,dataset  & dset, const dataset  & dset2 , const options &Options );


  class randomGenerator
    {
        public:
            randomGenerator()
            {
                srand(static_cast<int>(time(0)));
            }
            int operator()(int remaining) const
            {
                return (rand() % remaining);
            }
    };


class profilet
{
 public:
  vector<kind> kinds;//tipus (enumerats)
  vector<int> types;// tipus (integer indica tipus)
  vector<int> mvalues;// int que indica valors diferents
};

class example
{
 public:
  byte *v;

  example(const long Size) {
    v= new byte [Size];
  }

  /*  NO ES POT FER PERQUE NO GUARDEM LA LONGITUD
 example( const example & e) {
   v= new (byte) [
 }
 */

 ~example() {
 /* NO BORRA PERQUE NO TENIM CONST. COPIA I DONARIA PROBLEMES. ES BORREN EN EL DESTRUCTOR DE DATASET
    if(v!=NULL)	
      delete [] v; */
 }
};

class dataset {

  public:
  vector<example> data; //vector de exemples
  word num_arguments; //num de arguments
  sizeexamplet size_example;// tamany en bytes de cada exemple
  vector<word> argoffset;// vector de offset per accedir als arguments
  exetablet exetable;// taula amb els literals dels arguments
  profilet profile;// perfil de la funcio
  vdistt vdist; // distribuci de classes
  vector<string> names; // names of the arguments and classes optional
  vector<numeric> weights;// weights of the values of classes
  discrete Majority_Class; // Majority class of dataset

  vector<pair<string,string> > class_equivalence;


  dataset()
    {
      num_arguments=0;
      size_example=0;
    }


  dataset(const string & fitxer, const options &Options)
    {
      num_arguments=0;
      size_example=0;
      ompli_memoria(fitxer, (*this), Options);
      calcula_majorityclass();
    }
  ~dataset()
  {
   //delete argoffset ;
    for (indext i=0;i<data.size();i++)
     	delete [] data[i].v;   // FA FALTA PERQU example no es destruix correctament
   }

  dataset(const string &fitxer,const dataset &dtrain, const options &Options):vdist(dtrain.NumClasses())
    {
      num_arguments=0;
      size_example=0;
    if (!fitxer.empty())  ompli_memoria(fitxer, (*this), dtrain, Options);// per si de cas no li hem passat fitxer
    calcula_majorityclass();
    }

  dataset(const dataset &dtrain, const unsigned long NExamples, const options &Options); // generates a new random dataset without touching the dtrain


  void Append(const dataset &Second);  // appends the examples of the second dataset to the first dataset
                                  // the data is copied completely (and deeply), and the original dataset unaffected.

  void show()
    {

      int j;
      word4 i;
      for (i=0;i<data.size();i++)
	{
	  example &ex =data[i];
	  printf("\n");
	  for (j=0;j<num_arguments;j++)
	    {
	      if ((profile.kinds[j]==NUMERIC_KIND) || (profile.kinds[j]==NUMERIC_WITH_UNKNOWN_KIND))
 		printf(",%f",*((numeric *)(ex.v+argoffset[j])));
 		    else if (profile.kinds[j]==IGNORE_KIND)
 		      printf(",IGNORED");
	      else
		printf(",%d",ex.v[argoffset[j]]);
	    }
	}
    }

  void OutputOnFile(const char *File) const {
    FILE *F= NULL;
    OutputOnFile2(File, true, true, F);
  }


  void OutputOnFile2(const char *File, bool Open, bool Close, FILE * &F) const;


  static void  OutputTwoDatasetsOnFile(const char *OutputFile2, const dataset *Train, const dataset *Invented) {
    FILE *F;
    Train->OutputOnFile2(OutputFile2, true, false, F);
    Invented->OutputOnFile2(OutputFile2, false, true, F);
  }


    /* mostra la distrubuci de les classes*/
  void show_vdist()
  {
  cout<<" ";
  int i;
   for(i=0; i<NumClasses(); i++)
	  cout<<i<<"    ";
	  cout<<endl;
	for(i=0; i<NumClasses(); i++)
	     cout<<" "<< vdist[i];
	      cout<<endl;
	for(i=0; i<NumClasses(); i++)
	     cout<<" "<< (float)vdist[i]/Cardinality()<<"%";
	      cout<<endl;
  }
  
  inline argument Nth_arg(indext nex ,int  narg) const {
    argument Arg;
    kind Kind;
    
    const example &ex=data[nex];
    Kind=profile.kinds[narg];

    if ((Kind == NUMERIC_KIND) || (Kind == NUMERIC_WITH_UNKNOWN_KIND))
      Arg.Value_n= *(numeric *)(ex.v + argoffset[narg]);
    else if (Kind == DISCRETE_KIND) 
      Arg.Value_d= *(discrete *)(ex.v + argoffset[narg]);
    else if (Kind == VARIABLE_KIND) 
      Arg.Value_v= *(variable *)(ex.v + argoffset[narg]);
    else if (Kind == IGNORE_KIND) {
      // NOTHING TO DO
    }
    else cout << "Error: unknown argument kind in function Nth_arg";
    return Arg;
  }

 inline void Set_Nth_arg(example &ex ,int narg, argument Arg) {
    kind Kind;

    Kind=profile.kinds[narg];

    if ((Kind == NUMERIC_KIND) || (Kind == NUMERIC_WITH_UNKNOWN_KIND))
      *(numeric *)(ex.v + argoffset[narg])= Arg.Value_n;
    else if (Kind == DISCRETE_KIND)
      *(discrete *)(ex.v + argoffset[narg])= Arg.Value_d;
    else if (Kind == VARIABLE_KIND)
      *(variable *)(ex.v + argoffset[narg])= Arg.Value_v;
    else if (Kind == IGNORE_KIND) {
      // NOTHING TO DO
    }
    else cout << "Error: unknown argument kind in function Nth_arg";
  }

  inline long Number_Of_Numeric() const {
    kind Kind;
    long cont=0;

    for(long narg=0; narg < num_arguments - 1; narg++) {  // num_arguments -1 to avoid counting the class
      Kind=profile.kinds[narg];

      if ((Kind == NUMERIC_KIND) || (Kind == NUMERIC_WITH_UNKNOWN_KIND))
        cont++;
     }
    return cont;
  }


  inline long Number_Of_Nominal() const {
    kind Kind;
    long cont=0;

    for(long narg=0; narg < num_arguments - 1; narg++) {  // num_arguments -1 to avoid counting the class
      Kind=profile.kinds[narg];

      if (Kind == DISCRETE_KIND)
        cont++;
     }
    return cont;
  }

  inline void OrderedSetOfRealsOf(int narg, set<numeric> &s, const vector<indext> &Covered)  const {
    vector<indext>::const_iterator i=Covered.begin(), n=Covered.end();
    for(; i<n; i++) {
      s.insert( *(numeric *)(data[*i].v + argoffset[narg]) ) ;          // Ordered insertion without duplicates
    }
  }

  
  inline bool FollowsCondition(int i, const condition & c, indext e) const{
    int a;

    numeric n= Nth_arg(e,i).Value_n;

    switch (c.Operator) {
      case CONST_EQ:     if ((Nth_arg(e,i)).Value_d != c.Value.Value_d) return false;
                                       break;
      case CONST_NEQ:  if ((Nth_arg(e,i)).Value_d == c.Value.Value_d) return false;
                                       break;
      case VAR_EQ:          a= (int)c.Value.Value_v;  // numero de l'altra variable
                                        if (Nth_arg(e,i).Value_d != Nth_arg(e,a).Value_d) return false;
                                        break;
      case VAR_NEQ:        a= (int)c.Value.Value_v;  // numero de l'altra variable
                                        if (Nth_arg(e,i).Value_d ==Nth_arg(e,a).Value_d) return false;
                                        break;
      case NUM_LT:          if (Nth_arg(e,i).Value_n >= c.Value.Value_n) return false;
                                        break;
      case NUM_GE:         if (Nth_arg(e,i).Value_n < c.Value.Value_n) return false;
                                        break;
      case NUM_LT_WU:      if ((Nth_arg(e,i).Value_n >= c.Value.Value_n) || (Nth_arg(e,i).Value_n == MISSING_NUMERIC)) return false;
                                        break;
      case NUM_GE_WU:      if ((Nth_arg(e,i).Value_n < c.Value.Value_n) || (Nth_arg(e,i).Value_n == MISSING_NUMERIC)) return false;
                                        break;
      case NUM_MISSING:    if (Nth_arg(e,i).Value_n != MISSING_NUMERIC)
                             return false;
                           n=n; break;
    // empty:    
    // default:   // res que fer
    }
    return true;
  }

  inline void Obtain_Results(vector<discrete> &results)
  {
   word4 i;
  for (i=0;i<data.size();i++)
     results.push_back(ClassOf(i));
     }

  inline discrete ClassOf(indext i) const
    {
      return data[i].v[argoffset[num_arguments-1]];
    }


  inline void AssignClass(indext k, discrete Pred) {
    data[k].v[argoffset[num_arguments-1]] = Pred;
  }

  inline kind KindOf(int a ) const
    {
      return profile.kinds[a];
    }

  inline int TypeOf(int a ) const
    {
      return profile.types[a];
    }

  inline int NumArguments() const   // returns the number of arguments not considering the class
    {
      return num_arguments - 1;
    }
  
  inline discrete NumConstants( int a) const
    {
      return exetable[a].size();
    }
 
  inline discrete NumClasses() const
    {
      return exetable[num_arguments - 1].size();
    }
  
  inline bool Covers(const rule & r, indext e) const
    {
      for(int a=0; a < r.Num_arg; a++) {
         const vector<condition> &CondVector= r.V_condition[a];
         int NumCond= CondVector.size();
         if (NumCond >= 0) {
           for (int i=0; i<NumCond; i++) {
	         if (!FollowsCondition(a, CondVector[i], e)) return false;
           }
         }
      }
      return true;
      /* Igual hauria de tornar tres valors:
	 no quadren els arguments
	 quadren els arguments pero no la classe
	 quadren els arguments i la classe */
    }

	inline indext Cardinality() const
		{
			return data.size();
		}
	
	 inline vdistt ClassDistribution()  const
	 	{
	 		return vdist;
	 	}
	 inline discrete MajorityClass() const
	   {
	     return Majority_Class;
	   }
	 
	 inline void calcula_majorityclass()  {
	   
	   indext Max= 0;
	   for(int i=0; i<NumClasses(); i++) {
	     indext iValue= vdist[i];
	     if (iValue > Max) {
	       Max= iValue;
	       Majority_Class= i;
	     }
	   }
	}
     inline float FreqOfClass(int Class) const  {
       return vdist[Class] / float(data.size());
     }

     inline string NameArgumentValue(int a, discrete v) const {
        return (exetable[a])[v];
     }

     inline string NameClassValue(discrete v) const {
        return (exetable[num_arguments - 1])[v];
     }


  // Name of the argument (e.g. "temperature"). Don't confuse with NameArgumentValue, which is the name of each of the values ("high", "low", ...)
	inline string GetName(discrete v) const {
		return names[v];
		}
		
	inline numeric GetWeight(discrete v) const {
		return weights[v];
		}

    inline void SetWeightsTo1() {
       if (weights.size() != 0)  {
         for(int i= 0; i < NumClasses(); i++) {
           weights[i]= 1.0;
         }
      }
      else {
          for(int i= 0; i < NumClasses(); i++) {
            weights.push_back(1.0);
          }
      }
    }

    inline void SetWeightsToInvFreq(float Param) {
       SetWeightsTo1(); // to initialise in case it's empty
       for(int i= 0; i < NumClasses(); i++) {
         float Dist= vdist[i];
         if (Dist != 0) {
           weights[i]= Param / Dist;
         }
         else {
           cout << "Function SetWeightsToInvFreq. Class with no elements in dataset" << endl;
         }
      }
    }

		
		// recupera l'index a partir de l'argument i del string
		// si no existeix, mostra missatge de error i aborta
  	inline discrete GetIndex(word numarg,char * arg, char *exemple) const {
	  	unsigned int cont=1;
 		  const argtable *taula;
  	  if (numarg>exetable.size())	{
  		  cout<<" \n ERROR incorrect number of arguments in test data ";
  		  exit(-1);
  	  }
  	  else {
  		  taula=&(exetable[numarg]);
  		  for (cont=0;cont<(*taula).size();) {
	        if ((*taula)[cont]==arg)
		        return cont;
     		  cont++;
  	  	}
		  	cout<<" \n ERROR unknown symbol :\"" << arg << "\" in argument " << numarg << " of exemple:" << endl;
		  	cout << "\"" << exemple << "\"" << endl;
  	  	exit(-1);
		  }
			
		  return 0; // never arrives here.
	  }

				
		// extrau un tant per cent (pcn) del exemples de data al atzar,  afegint-los
		// a un data set nou que es passa per parametre
		// si KFold == 0 desordena el vector i agafa el primer percentatge.
		// si KFold > 0 ja no torna a desordenar el vector i agafa el tros KFold
  void RandomExtract(numeric pcn, dataset  & nou, unsigned long KFold= 0) {

	  nou.num_arguments=(*this).num_arguments;
    nou.size_example=(*this).size_example;
 	  nou.argoffset=(*this).argoffset;
	  nou.exetable=(*this).exetable;// taula amb els literals dels arguments
    nou.profile=(*this).profile;// perfil de la funcio
    nou.vdist=(*this).vdist; // distribuci de classes
    nou.names=(*this).names; // names of the arguments and classe optional
    nou.weights=(*this).weights;// weights of the values of classes
  		
 //   if (KFold == 0) {
		  random_shuffle(data.begin(),data.end()); // desordenem el vector al atzar
 //   }

    if ((pcn>1)&&(pcn<=0)) {
		  cout << "ERROR: incorrect percentage in method RandomExtract of class dataset" << endl;
      exit(0);
	  }
    unsigned long arg =(unsigned long)(data.size()*pcn);
    unsigned long First, Last;
    First= arg * KFold;
    Last= arg * (KFold+1);
    if (Last > data.size()) {
      cout << "ERROR: incorrect KFold value in method RandomExtract of class dataset" << endl;
      exit(0);
	  }
    nou.data=vector<example> (&data[First], &data[Last]); //atenci copia del 0 al arg-1 per a KFold==0 i els segents blocs per a KFold>0
      data.erase((vector<example>::iterator)&data[First], (vector<example>::iterator)&data[Last]);

 	  nou.Setvdist(); // actualitza la distribuci de classes
 	  Setvdist();// actualitza la distribuci de classes
  }
	
			// extrau un tant per cent (pcn) del exemples de data al atzar,  afegint-los
		// a un data set nou que es passa per parametre, a mes  cada volta canvia
		// els elements que s'extrauen ja que es depen de la cridada times()
		// si KFold == 0 desordena el vector i agafa el primer percentatge.
		// si KFold > 0 ja no torna a desordenar el vector i agafa el tros KFold
  void RandomExtractr(numeric pcn, dataset  & nou, unsigned long KFold= 0) {

	  nou.num_arguments=(*this).num_arguments;
 	  nou.size_example=(*this).size_example;
 	  nou.argoffset=(*this).argoffset;
	    nou.exetable=(*this).exetable;
  		nou.profile=(*this).profile;// perfil de la funcio
  		nou.vdist=(*this).vdist; // distribuci de classes
  		nou.names=(*this).names; // names of the arguments and classe optional
  		nou.weights=(*this).weights;
 		 randomGenerator rg;
  			
//    if (KFold == 0) {
     	random_shuffle(data.begin(),data.end(),rg); // desordenem el vector al atzar
                                        //   ^^ ATTENTION TO "rg" (random generator)
//  	}

    if ((pcn>1)&&(pcn<=0)) {
		  cout << "ERROR: incorrect percentage in method RandomExtract of class dataset" << endl;
      exit(0);
	  }
    unsigned long arg =(unsigned long)(data.size()*pcn);
    unsigned long First, Last;
    First= arg * KFold;
    Last= arg * (KFold+1);
    if (Last > data.size()) {
      cout << "ERROR: incorrect KFold value in method RandomExtract of class dataset" << endl;
      exit(0);
	  }
    nou.data=vector<example> (&data[First], &data[Last]); //atenci copia del 0 al arg-1 per a KFold==0 i els segents blocs per a KFold>0
nou.data=vector<example> (&data[First], &data[Last]); //atenci copia del 0 al arg-1 per a KFold==0 i els segents blocs per a KFold>0
 	  nou.Setvdist(); // actualitza la distribuci de classes
 	  Setvdist();// actualitza la distribuci de classes
 	}	
		
// updates class distribution after any modification on the data
// and the majorityclass
  void Setvdist() {
    unsigned long i;
    int j=argoffset[num_arguments-1];
	  for (i=0;i<vdist.size();i++)
		  vdist[i]=0; // fiquem a 0 totes les classes
		
	  for (i=0;i<data.size();i++)
	    vdist[data[i].v[j]]++;  // actualitzem la distribuci de la classe
	
	  calcula_majorityclass();
	}
};




#endif



// Cardinality retorna indext

// ClassDistribution retorna vecotr<indext>
