/***************************************************************************
                          pmml.cpp -  description
                             -------------------

    functionality        : defines basics of pmml
    relevant classes     :
    relevant functions   : Predicate, TreeModel

    begin                : 26 Dic 2002
    copyright            : (C) 2002 by Ricardo Blanco Vega
    email                : rblanco@uach.mx
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "pmml.h"


 void pmml::Predicate(rule &Rule, char * FunctionName, bool Short,const char *Clase) {
    int i;
    char Buf[2000];
    strstream s(Buf, 2000);
    int NumVar= 0;
    int NumCond;
    int Num_arg= Rule.Get_Num_Arg();
    s.clear();
    int TotaCond=0;
    for(i=0; i < Num_arg-1; i++) {
      vector<condition> &CondVector= Rule.V_condition[i];
      TotaCond+=CondVector.size();
    }

    if (TotaCond>1)
       s << "           <CompoundPredicate booleanOperator=\"and\">"<<endl;



    NumVar= 0;
	for(i=0; i < Num_arg-1; i++) {
      vector<condition> &CondVector= Rule.V_condition[i];
      NumCond= CondVector.size();
      if (NumCond== 0) {
        // do nothing
      }
      else {

      for (int j=0; j<NumCond; j++) {
        discrete v;
        if ((CondVector[j].Operator) == CONST_EQ) {
            s << "             <SimplePredicate field=\"X" << NumVar << "\" operator=\"equal\" value=\"" ;
          v= (int)(CondVector[j].Value).Value_d;
          if (Short== true)
            s << v<< "\" />"<<endl;
          else {
            s << TrainSet.NameArgumentValue(i, v)<< "\" />"<<endl ;
          }
        }
        else if ((CondVector[j].Operator) == CONST_NEQ) {
          s << "             <SimplePredicate field=\"X" << NumVar << "\" operator=\"notEqual\" value=\"" ;
          v= (int)(CondVector[j].Value).Value_d;
          if (Short== true)
            s << v << "\" />"<<endl;
          else {
            s << TrainSet.NameArgumentValue(i, v) << "\" />"<<endl ;
          }
        }
        else if ((CondVector[j].Operator) == VAR_EQ) {
            s << "             <SimplePredicate field=\"X" << NumVar << "\" operator=\"equal\" value=\"X"
              <<(int)(CondVector[j].Value).Value_v << "\" />"<<endl ;
        }
        else if ((CondVector[j].Operator) == VAR_NEQ) {
            s << "             <SimplePredicate field=\"X" << NumVar << "\" operator=\"notEqual\" value=\"X"
              <<(int)(CondVector[j].Value).Value_v << "\" />"<<endl ;
        }
        else if (((CondVector[j].Operator) == NUM_LT) || ((CondVector[j].Operator) == NUM_LT_WU) ) {
            s << "             <SimplePredicate field=\"X" << NumVar << "\" operator=\"lessThan\" value=\"X"
              <<(int)(CondVector[j].Value).Value_n << "\" />"<<endl ;
        }
        else if (((CondVector[j].Operator) == NUM_GE) || ((CondVector[j].Operator) == NUM_GE_WU)) {
            s << "            <SimplePredicate field=\"X" << NumVar << "\" operator=\"greaterOrEqual\" value=\""
              <<(CondVector[j].Value).Value_n << "\" />"<<endl ;
        }
        else if ((CondVector[j].Operator) == NUM_MISSING) {
            s << "            <SimplePredicate field=\"X" << NumVar << "\" operator=\"equal\" value=\"?\" />"<<endl ;
        }
        else
          cout << "Error in function pmml::Predicate: Operator unknow (no conocido).";
      }
      }
      NumVar++;
    }

  if (TotaCond>1)
       s <<"           </CompoundPredicate>"<<endl;

       s <<"           <Node score=\""<<Clase<<"\"><True></Node>"<<endl;

    s << '\0';  // end of string

    fputs(Buf,Out);
  }

  // crea el modelo de rbol que cumpla con el dtd de PMML a partir de una solucion (set of leaf nodes)
  void pmml::TreeModel(char *FilePmml,const vector<and_tree *> Sol, char *FunctionName)
  {
    char cade[800];
    int i;
     string Clase;

     if ((Out=fopen(FilePmml,"wt"))==NULL)	{
       cout << "Error when creating file : \"" << FilePmml << "\"" << endl;
       exit(-1);
     }

//-- put header.

  fputs("<?xml version=\"1.0\" ?>\n",Out);
  fputs("<PMML version=\"2.0\">\n",Out);
  fputs("<Header copyright=\"www.dsic.upv.es\"\n",Out);
  fputs("        description=\"SMILES 2.3.6\"/>\n",Out);

// DataDictionary
  sprintf(cade,"   <DataDictionary numberOfFields=\"%i\" >\n",TrainSet.num_arguments );
  fputs(cade,Out);
  for (i=0; i<TrainSet.num_arguments; i++)
  {
      if (TrainSet.KindOf(i)==DISCRETE_KIND)
      {
          if(TrainSet.names.size()!=0)
              sprintf(cade,"      <DataField name=\"%s\" optype=\"categorical\" >\n",TrainSet.names[i].c_str() );
          else
              sprintf(cade,"      <DataField name=\"X%i\" optype=\"categorical\" >\n",i );
          fputs(cade,Out);

          for (int v=0; v<TrainSet.exetable[i].size() ;v++)
          {
            sprintf(cade,"         <Value value=\"%s\"/>\n",TrainSet.NameArgumentValue(i, v).c_str() );
            fputs(cade,Out);
          }

          fputs("      </DataField>\n",Out);
      }
      else
      {
          if(TrainSet.names.size()!=0)
              sprintf(cade,"      <DataField name=\"%s\" optype=\"continuous\" />\n",TrainSet.names[i].c_str() );
          else
              sprintf(cade,"      <DataField name=\"X%i\" optype=\"continuous\" />\n",i );
          fputs(cade,Out);
      }
  }
  fputs("   </DataDictionary>\n",Out);

// TreeModel and MiningSchema
    char Buf[800];
    strstream s(Buf, 800);
    int NumVar= 0;
    int Num_arg=Sol[0]->Rule.Get_Num_Arg();
    s.clear();
    s << "   <TreeModel modelName=\""<<FilePmml<< "\">"<<endl;
    s << "      <MiningSchema>" << endl;

    for(i=0; i < Num_arg; i++) {
      s <<"          <MiningField name=\"X" << NumVar<<"\"";
      NumVar++;
      if (i!=Num_arg-1) s << "/>" << endl;
    }
    s << " usageType=\"predicated\"/>" << endl;

    s << "      </MiningSchema>" << endl;

     discrete Cl= Sol[0]->Get_PredictedClass();
     Clase=TrainSet.NameClassValue(Cl);
    strcpy(Tab,"         ");
    s << "     <Node score=\""<<Clase<<"\">"<<endl;
    s << '\0';
    fputs(Buf,Out);

// Nodes
    long Size= Sol.size();
    for (int i=0; i<Size; i++) {
      discrete Cl= Sol[i]->Get_PredictedClass();
      sprintf(cade,"%s<Node score=\"%s\">\n",Tab,Clase.c_str());
      if (i!=0)
        fputs(cade,Out);
      Predicate(Sol[i]->Rule, FunctionName, false,TrainSet.NameClassValue(Cl).c_str());
      sprintf(cade,"%s</Node>\n",Tab);
      if (i!=0)
        fputs(cade,Out);
    }

// to put tags of lock finally in the file.
   fputs("      </Node>\n",Out);
   fputs("   </TreeModel>\n",Out);
   fputs("</PMML>\n",Out);

   fclose(Out);
}
